/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



#ifndef MP7JRSEXCEPTION_H
#define MP7JRSEXCEPTION_H

#include <xercesc/util/XMLString.hpp>
#include "Mp7JrsDefines.h"

typedef enum {
	MP7JRS_WARNING,
	MP7JRS_ERROR,
	MP7JRS_FATAL_ERROR
} Mp7JrsExSeverity;

typedef enum {
	MP7JRS_AUTODELETE, // Message created on stack, nothing to do
	MP7JRS_NOAUTODELETE, // Message created on our heap (NOT Xerces heap), we have to destroy it.
} Mp7JrsMsgMemory;

/** Primary exception class. */
class MP7JRS_EXPORT Mp7JrsException 
{
public:
	/**
	 * This is version 2.0
	 */
	Mp7JrsException(Mp7JrsExSeverity sev, const XMLCh * msg, Mp7JrsMsgMemory mode = MP7JRS_AUTODELETE) 
		: _severity(sev) { 
		_msg = XMLString::transcode(msg); // Xerces heap
		if(mode == MP7JRS_NOAUTODELETE && msg){
			XMLCh * tmp = const_cast<XMLCh*>(msg);
			delete [] tmp; // Our heap
		}		
	}

	/**
	 * This is version 2.0
	 */
	Mp7JrsException(Mp7JrsExSeverity sev, const char * msg, Mp7JrsMsgMemory mode = MP7JRS_AUTODELETE) 
		: _severity(sev) { 
		_msg = XMLString::replicate(msg);  // Xerces heap
		if(mode == MP7JRS_NOAUTODELETE && msg){
			char * tmp = const_cast<char*>(msg);
			delete [] tmp; // Our heap
		}
	}

	// Copy ctor, otherwise we leave a memory leak!
	Mp7JrsException(const Mp7JrsException & other)
	{
		_severity = other._severity;
		_msg = XMLString::replicate(other._msg);  // Xerces heap
	}
	
	virtual ~Mp7JrsException()
	{
		XMLString::release(&_msg); // Xerces heap, _msg is reset to NULL
	}
	
	/** Get severity identifier */
	Mp7JrsExSeverity GetSeverity() { return _severity; };

	/** Get the severity as string. */
	const char* GetSeverityText() {
		switch (_severity) {
		case MP7JRS_WARNING: return "Warning"; 
		case MP7JRS_ERROR: return "Error";
		case MP7JRS_FATAL_ERROR: return "Fatal error";
		default: return "";
		};
	}

	/** Get a descriptive error message. 
	  * Since the original GetMessage() method is overloaded 
	  * in Windows with a Microsoft macro and expands to either
	  * GetMessageA or GetMessageW, we change our method to avoid this.
	  */
	const char* GetMessageText() { return _msg; }

private:
	Mp7JrsExSeverity _severity;
	char *_msg;
};

#endif
