/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


#ifndef MP7JRSFACTORY_H
#define MP7JRSFACTORY_H

#include "Mp7JrsDefines.h"

#include "Mp7JrsPtr.h"
#include "Mp7JrsPtrTypes.h"

#include "Mp7JrsException.h"

typedef const char * SimpleGuid; // Extensible library version with real guids. Note, the const is to make GCC happy

class MP7JRS_EXPORT Mp7JrsLicenseException : public Mp7JrsException 
{
	public:
	Mp7JrsLicenseException(const char* msg) : Mp7JrsException(MP7JRS_ERROR,msg) {}  
};

/** Core factory base class. */
class MP7JRS_EXPORT Mp7JrsFactory
{

public:
	/** @name Initialization
	 */
	//@{
	/** Initialize factory registry. */
	static void Initialize();
	
	/** Clean up factory registry. */
	static void Terminate();

	/** Returns the version string of the library. This normally changes, if the XML schema has changed. */
	static const char * GetVersion();
	
	/** Returns the version string of the source code design model for this library. This normally changes, if a template has been adjusted. */
	static const char * GetModelVersion();
	
	//@}

	/** Map the possibly unordered type id to an ordered index. Note: this is used for registering only! */
	static unsigned int MapGuidToIndex(SimpleGuid id);

	/** Create a node object with type given as index */
	static Mp7JrsNodePtr CreateObject(unsigned int index);
	
	/** Clone the given node */
	static Mp7JrsNodePtr CloneObject(const Mp7JrsNodePtr &original);

	/** Get the classname associated with the given factory index. */
	static const char * GetTypeName(unsigned int index);
	/** Get the factory index (i.e. type id) for a certain classname. */
	static unsigned int GetTypeIndex(const XMLCh * qname);

	/** Delete the given object. */
	static void DeleteObject(Mp7JrsNode * item);

	/** "Delete" the smart pointer (nothing to do) */
	template<class P>
	static void DeleteObject(Mp7JrsPtr<P> & /*item*/) {}

	/** Set if the factory must be deleted when cleaning up the factory registry */
	void SetToDelete(bool todelete) {m_ToDelete = todelete;}

	/** Get information, if factory must be deleted when factory registry is cleaned up. */
	bool GetToDelete() const {return m_ToDelete;}

	// FTT For binary compatibility
	static void SetNamespaceUri(const XMLCh * prefix, const XMLCh * nsuri);
	// FTT For binary compatibility
	static const XMLCh * GetNamespaceUri(const XMLCh * prefix);
	// FTT For binary compatibility
	static void UnregisterFactory(SimpleGuid id);

protected:
	Mp7JrsFactory(SimpleGuid id, const char * classname);
	virtual ~Mp7JrsFactory(void); // Don't use delete, use DeleteFactory() instead;

	virtual void DeleteFactory();

	virtual Mp7JrsNodePtr CreateObject() = 0;
	virtual const char * GetTypeName() const = 0;
	virtual void DeleteRawObject(Mp7JrsNode * item);

private: 
	static int initialized;
	
	// Unimplemented methods
	Mp7JrsFactory(const Mp7JrsFactory &);

protected:
	// False, if the factory is created on stack (default)
	// Must be set to true when created on heap.
	bool m_ToDelete;

	// We must know our ordered class id for adding it to our generated nodes
	unsigned int m_ClassId;
	
};

#endif // MP7JRSFACTORY_H
