/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



#ifndef MP7JRS_NODE_H
#define MP7JRS_NODE_H

// Intro of API documentation
#include "docintro.h"

#include "Mp7JrsDefines.h"
#include "Mp7JrsPtr.h"
#include "Mp7JrsPtrTypes.h"
#include "Mp7JrsEnumerator.h"

#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/sax/InputSource.hpp>
#include <xercesc/framework/XMLFormatter.hpp>

// This macro has been removed in version 2.0.
// If you need this macro, please do it in your own source code!
// Method GetClassType has been renamed to GetTypeName()
// #define GetClassType GetTypeName


// The baseclass for all factories
class Mp7JrsFactory;

// Helperclass for XPath functions
struct Mp7JrsXPathParseContext;

typedef enum {
	MP7JRS_NOCREATE, // Get existing nodes or nothing.
	MP7JRS_FORCECREATE, // Create missing nodes
} Mp7JrsXPathCreation; // Must come before 


/** Basic Mp7Jrs node type. */
class MP7JRS_EXPORT Mp7JrsNode
{

friend class Mp7JrsFactory;

public:
	/** Get class type name as string. */
	const char * GetTypeName() const;

	/** Get the class ID of this node (i.e. the MPEG-7 type it represents. */
	const unsigned int GetClassId() const;

	/** Set the content name of this node. */
	void SetContentName(XMLCh * contentname);

	/** Get the content name of this node */
	const XMLCh * GetContentName() const;

	/** Convert the content to string.  Note, this is purely for ComplexTypes with SimpleContent and SimpleTypes. All other types return the class name.
	  * @return The converted content of this class or NULL if something goes terribly wrong. */
	virtual XMLCh * ToText() const = 0;

	/** Parse the content of this type. Note, this is purely for ComplexTypes with SimpleContent and SimpleTypes. All other types return false.
	  * @param txt buffer to parse from.
	  * @return True if parsing was successful, false otherwise..
	  */
	virtual bool Parse(const XMLCh * const txt) = 0;

	/** Parse the content from a string. Note, this is purely for ComplexTypes with SimpleContent and SimpleTypes.
	  * @param txt buffer to parse from.
	  * @return This default implementation always return false.
	  */
	virtual bool ContentFromString(const XMLCh * const txt);

	/* Convert the content to string.  Note, this is purely for ComplexTypes with SimpleContent and SimpleTypes. 
	 	 * @return This default implementation always return NULL.
	 	 */
	virtual XMLCh * ContentToString() const;

	/** Serialialize this object to a DOM tree.
	  * @param doc DOM document.
	  * @param parent Parent node in DOM tree.
	  */
	virtual void Serialize(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument * doc, XERCES_CPP_NAMESPACE_QUALIFIER DOMElement* parent, XERCES_CPP_NAMESPACE_QUALIFIER DOMElement** newElem = NULL) = 0;

	/** Deserialize this object from a DOM tree.
	  * @param doc DOM document.
	  * @param parent Parent element in DOM tree.
	  * @param current Current element in DOM tree.
	  */
	virtual bool Deserialize(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument * doc, XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * parent, XERCES_CPP_NAMESPACE_QUALIFIER DOMElement ** current) = 0;

	/* Creates and inserts a childnode of the specified type
	 * @deprecated In order to reduce API size (KISS), this method will be removed in the next version.
	 * @param elementname The tag type of the child to be created (XML element name).
	 * @param xsitype The xsi type of the child to be created if available.
	 * @return the created child with the correct dynamic type for the specified type.
	 */
	virtual Mp7JrsNodePtr CreateChild(const XMLCh * const elementname, const XMLCh * const xsitype);

	/** Get the absolute XPath expression describing this node.
	  * This method is experimental...
	  * @return An absolute XPath expression describing this node.
	  */
	virtual XMLCh *GetXPath() const;

	/** Get the node represented by this xpath.
	  * @deprecated In order to reduce API size (KISS), this method will be removed in the next version.
	  * @return A node pointer or NULL if an invalid xpath argument was given.
	  * NOTE: This method implements only a very restricted subset of
	  * XPath: only the child path is supported, only a single node is
	  * returned.
	  */
	Mp7JrsNodePtr NodeFromXPath(const char *xpath) const;
	virtual Mp7JrsNodePtr NodeFromXPath(const XMLCh *xpath) const;

	/** Get the node(s) represented by this xpath.
	  * Missing nodes are created and inserted/appended automatically.
	  * Note, we don't fill in missing elements in between.
	  * If a collection contains 2 elements and you are requesting the 4th, then an exception is thrown.
	  * @return An enumeration of nodes, which may be empty.
	  * NOTE: This method implements only a very restricted subset of
	  * XPath: xsi:type attributes are allowed, and the order of 
	  * collection indexes and xsi:types is taken into account.
	  * 
	  * For more information, please see the class documentation
	  */
	Mp7JrsNodeEnum GetOrCreateFromXPath(const char *xpath, Mp7JrsClientID client = MP7JRS_UNDEFINED_CLIENT, Mp7JrsXPathCreation flag = MP7JRS_FORCECREATE);
	virtual Mp7JrsNodeEnum GetOrCreateFromXPath(const XMLCh *xpath, Mp7JrsClientID client = MP7JRS_UNDEFINED_CLIENT, Mp7JrsXPathCreation flag = MP7JRS_FORCECREATE, Mp7JrsXPathParseContext * context = NULL);
	
	/**
	  * The same like GetOrCreateFromXPath() just read only.
	  * For more information, please see the class documentations of GetOrCreateFromXPath.
	  */
	Mp7JrsNodeEnum GetFromXPath(const char *xpath ) /* const */;
	/**
	  * The same like GetOrCreateFromXPath() just read only.
	  * For more information, please see the class documentations of GetOrCreateFromXPath.
	  */
	Mp7JrsNodeEnum GetFromXPath(const XMLCh *xpath) /* const */;


protected:
	// Consume leading "..", go upward in tree, then continue with
	// xpath from there.
	Mp7JrsNodePtr UpwardWithXPath(const XMLCh *xpath) const;
	
public:

	/** Get the parent of the node.
	  * @return A pointer to the parent of the node or NULL if this is the root node of a tree fragment.
	  */
	virtual Mp7JrsNodePtr GetParent() const;

	/** @name Locking
	 */
	//@{

	/** Request to lock this node for any operation.
	  * @param id ID of the client that wants to lock this node.
	  * @param receiveNotification Determines whether the client will be notifified later, in the case that the
	  *	lock cannot be granted immediately.
	  * @return True if the node has been locked, false otherwise.
	  * <b>NOTE: Not implemented, will currently always return true.</b>
	  */
	virtual bool Lock(Mp7JrsClientID id, bool receiveNotification = true);

	/** Request to lock this node for add or insert.
	  * @param id ID of the client that wants to lock this node.
	  * @param receiveNotification Determines whether the client will be notifified later, in the case that the
	  *	lock cannot be granted immediately.
	  * @return True if the node has been locked, false otherwise.
	  * <b>NOTE: Not implemented, will currently always return true.</b>
	  */
	virtual bool Lock4Add(Mp7JrsClientID id, bool receiveNotification = true);

	/** Unlock this node.
	  * @param id ID of the client that wants to unlock.
	  */
	virtual void Unlock(Mp7JrsClientID id);

	//@}

	/** @name Notifications
	 */
	//@{

	/** Register to receive notifications from this node.
	  * @param id ID of the client that wants to receive notifications.
	  * @param action Type of action, for which notifications shall be sent.
	  * @param subtree If true, the client is also registered to receive notifications from all nodes
	  * in the subtree below the specified node. This may be slow, if many nodes are involved.
	  * <b>NOTE: Not implemented. No notifications will be sent</b>
	  */
	virtual void Register(Mp7JrsClientID id, Mp7JrsNotificationAction action, bool subtree = false);

	/** Unregister from receiving notifications from this node.
	  * @param id ID of the client that wants to unregister.
	  * @param action Type of action, for which to unregister.
	  * @param subtree If true, the client is also unregistered from receving notifications from all nodes
	  * in the subtree below the specified node.
	  */
	virtual void Unregister(Mp7JrsClientID id, Mp7JrsNotificationAction action);

	//@}

	virtual ~Mp7JrsNode(void);

protected:
	Mp7JrsNode(void);
	virtual void DeepCopy(const Mp7JrsNodePtr &original) = 0;

	// checks if the specified client possesses a lock on this node
	virtual bool IsLocked(Mp7JrsClientID id);

	// FTT: deprecated In order to reduce API size (KISS), this method will be removed in the next version.
	// Returns all node element children (no attributes).  Only
	// applicable for nodes of type Class.
	virtual Mp7JrsNodeEnum GetAllChildElements() const = 0;

	Mp7JrsClientID lockOwner;

	// Overwritten in nodes of type Class to return the base of the
	// "inheritance" hierarchy
	virtual Mp7JrsNodePtr GetBaseRoot() { return m_myself.getPointer(); }

	virtual const Mp7JrsNodePtr GetBaseRoot() const { return m_myself.getPointer(); }

public:
	// public because it's called from unrelated-by-inheritance nodes
	// that happen to be the new parent of the node.
	void SetParent(const Mp7JrsNodePtr &new_parent);

private:
	// Unimplemented
	Mp7JrsNode(const Mp7JrsNode &);

	// Implemented in ClassTypes
	virtual void Init() {}

public:
	/** Set if xsi:type attribute shall be written when serialising this node (useful for elements having abstract types). */
	bool UseTypeAttribute;

protected:
	unsigned int m_ClassId;
	XMLCh * m_ContentName;

	Mp7JrsWeakPtr< Mp7JrsNode > m_Parent;

public:
	// What's this?  When storing any element, we need to do
	// m_element.setParent(this), only we can't use the `this' pointer
	// directly because we also need the reference count associated
	// with ourselves.  This is set by the factory during creation; if
	// the node is the base of another node, it's set to point to the
	// inheriter by inheriter's Init() method (can't be done by the
	// Factory, since m_Base isn't provided by the generic Node
	// class), so that all elements and attributes always point to the
	// "most specific parent".
	Mp7JrsWeakPtr< Mp7JrsNode > m_myself;
};

#endif // MP7JRSNODE_H
