/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



// Heavily inspired by the boost smart_ptr class 
// and Andrei Alexandrescu, "Modern C++ Design"

#ifndef MP7JRS_POINTER_INCLUDED
#define MP7JRS_POINTER_INCLUDED

#if defined(WIN32)
// FTT: To get rid of nasty 
// "c:\Programme\Microsoft Visual Studio .NET 2003\Vc7\PlatformSDK\Include\MsXml.h(9594): fatal error C1001: INTERNAL COMPILER ERROR"
#	ifndef WIN32_LEAN_AND_MEAN
#		define WIN32_LEAN_AND_MEAN
#	endif
#	include <windows.h>
#	pragma warning( disable : 4290 )
#elif defined(__APPLE__)
#	include <pthread.h>
#elif defined(__linux__)
#	include <pthread.h>
#else
#	error TODO: Add header for mutex type.
#endif


#include "Mp7JrsException.h"
#include "Mp7JrsUtil.h"

class Mp7JrsNode;

class MP7JRS_EXPORT Mp7JrsPtrException : public Mp7JrsException {

public:

	Mp7JrsPtrException(Mp7JrsExSeverity sev, const XMLCh* msg) 
		: Mp7JrsException(sev, msg) 
	{}

	Mp7JrsPtrException(Mp7JrsExSeverity sev, const char* msg) 
		: Mp7JrsException(sev, msg) 
	{}
};




class Mp7JrsNode;

class MP7JRS_EXPORT Mp7Jrscounter {

private:
#if defined(WIN32)
	CRITICAL_SECTION crit_sect_;
#elif defined(__APPLE__)
	pthread_mutex_t mutex_;
#elif defined(__linux__)
	pthread_mutex_t mutex_;
#else
#	error("TODO: Define Platform-specific mutex type.")
#endif

	// use count is always <= weak count.  When use count drops to
	// zero, the pointer can be deleted.  When weak count drops to
	// zero, counter self-destructs.
	long use_count_;
	long weak_count_;

private:
	// private: each counter is unique
	Mp7Jrscounter(Mp7Jrscounter const &);
	Mp7Jrscounter & operator=(Mp7Jrscounter const &);

public:
	Mp7Jrscounter(Mp7JrsNode *p) : use_count_(1), weak_count_(1), ptr(p)
	{
#if defined(WIN32)
		InitializeCriticalSection(&crit_sect_);
#elif defined(__APPLE__)
		pthread_mutex_init(&mutex_, NULL);	
#elif defined(__linux__)
		pthread_mutex_init(&mutex_, NULL);
#else
#	error("TODO: Define platform-specific mutex init func.")
#endif
	}

	~Mp7Jrscounter()
	{
#if defined(WIN32)
		DeleteCriticalSection(&crit_sect_);
#elif defined(__APPLE__)
		pthread_mutex_destroy(&mutex_);
#elif defined(__linux__)
		pthread_mutex_destroy(&mutex_);
#else
#	error("TODO: Define platform-specific mutex destroy func.")
#endif
	}

public:
	// Create a counter.  Don't do this via new in a method defined 
	// in this file, because it would be allocated in the heap of
	// whatever program that included the header.
	static Mp7Jrscounter *makeCounter(Mp7JrsNode *p);
	
	// baw 22 04 2004
	// Delete a counter, don't do this via delete in this file,
	// because it would be attempted to delete from the wrong heap.
	static void deleteCounter(Mp7Jrscounter* cntr);

protected:

	class _lock;
	friend class _lock;

	class _lock
	{
	private:

		Mp7Jrscounter & b_;

		_lock(_lock const &);
		_lock & operator=(_lock const &);

	public:

		explicit _lock(Mp7Jrscounter & b): b_(b)
		{
#if defined(WIN32)
			EnterCriticalSection(&b_.crit_sect_);
#elif defined(__APPLE__)
			pthread_mutex_lock(&(b_.mutex_));
#elif defined(__linux__)
			pthread_mutex_lock(&(b_.mutex_));
#else
#			error("TODO: Define platform-specific mutex lock func.")
#endif
		}

		~_lock()
		{
#if defined(WIN32)
			LeaveCriticalSection(&b_.crit_sect_);
#elif defined(__APPLE__)
			pthread_mutex_unlock(&(b_.mutex_));
#elif defined(__linux__)
			pthread_mutex_unlock(&(b_.mutex_));
#else
#			error("TODO: Define platform-specific mutex unlock func.")
#endif
		}
	};

	// Called when weak_count_ drops to zero
	void destruct()
	{
		// baw 22 04 2004: use static method of counter class for deleting it
		//Mp7JrsUtil::DeleteInDll(this);
		Mp7Jrscounter::deleteCounter(this);
	}

	// Called when use_count_ drops to zero; responsible for releasing
	// the managed pointer
	void dispose()
	{
		disposeUsingFactory(ptr);
	}

	// Helper function: non-template so it can be defined in a
	// cpp-file, defined in a cpp file so Mp7JrsFactory.h doesn't have
	// to be included here.
	void disposeUsingFactory(Mp7JrsNode *ptr);

public:

	void add_ref()
	{
		_lock lock(*this);
		if(use_count_ <= 0) throw Mp7JrsPtrException(MP7JRS_FATAL_ERROR, "Attempted to add reference to invalid pointer.");
		++use_count_;
		++weak_count_;
	}

	void release()
	{
		{
			_lock lock(*this);
			long new_use_count = --use_count_;

			if(new_use_count > 0)
			{
				--weak_count_;
				return;
			}
		}
		dispose();
		weak_release();
	}

	void weak_add_ref() // nothrow
	{
		_lock lock(*this);
		++weak_count_;
	}

	void weak_release() // nothrow
	{
		long new_weak_count;

		{
			_lock lock(*this);
			new_weak_count = --weak_count_;
		}

		if(new_weak_count == 0)
		{
			destruct();
		}
	}

	long use_count() const
	{
		_lock lock(const_cast<Mp7Jrscounter &>(*this));
		return use_count_;
	}

private:
	Mp7JrsNode *ptr;	// the managed pointer

};



template<class P>
class Mp7JrsWeakPtr;

/** Smart pointer template */
template<class P>
class Mp7JrsPtr {

public:

	Mp7JrsPtr()
		: ptr(0), use_count_(0)
	{}

	template<class W>
	explicit Mp7JrsPtr(const W *p)
	{
		if (!p) {
			ptr = 0;
			use_count_ = 0;
		} else {
			W *dummy = const_cast<W *>(p);
			ptr = dynamic_cast<P *>(dummy);
			if (ptr)
				use_count_ = Mp7Jrscounter::makeCounter(dynamic_cast<Mp7JrsNode *>(ptr));
			else {
				use_count_ = NULL;
#ifdef ASSIGNMENT_EXCEPTION
				// cast failed because of invalid assignment type
				throw Mp7JrsPtrException(MP7JRS_ERROR, "Invalid cast. Attempted to assign invalid pointer type.");
#endif
			}
		}
	}

	template<class W>
	Mp7JrsPtr(const Mp7JrsPtr<W>& p)
	{
		if (*(p.__count())) {
			(*p.__count())->add_ref();
			// It is now safe to access p.Ptr()
			ptr = dynamic_cast<P *>(p.Ptr());
			if (ptr) {
				use_count_ = const_cast<Mp7Jrscounter *>(*(p.__count()));
			}
			else {
				// Couldn't convert p.Ptr() to our type -> drop reference
				(*p.__count())->release();
				use_count_ = NULL;
#ifdef ASSIGNMENT_EXCEPTION
				// cast failed because of invalid assignment type
				throw Mp7JrsPtrException(MP7JRS_ERROR, "Invalid cast. Attempted to assign invalid pointer type.");
#endif
			}
		} else {
			// p is a null pointer, so are we
			ptr = NULL;
			use_count_ = NULL;
		}
	}

	// overwrite default copy constructor -- a template constructor is
	// never a default constructor, even if its argument would match.
	// VC++6 needs this after the template constructor
	Mp7JrsPtr(const Mp7JrsPtr<P>& p)
	{
		if (p.use_count_) {
			p.use_count_->add_ref();
			// It is now safe to access p.ptr
			ptr = p.ptr;
			use_count_ = p.use_count_;
		} else {
			// p is a null pointer, so are we
			ptr = NULL;
			use_count_ = NULL;
		}
	}

	explicit Mp7JrsPtr(const Mp7JrsWeakPtr<P>& p)
	{
		if (p.__count()) {
			(*p.__count())->add_ref();
			// It is now safe to access p.Ptr()
			ptr = dynamic_cast<P *>(p.Ptr());
			if (ptr) {
				use_count_ = const_cast<Mp7Jrscounter *>(*(p.__count()));
			}
			else {
				// Couldn't convert p.Ptr() to our type -> drop reference
				(*p.__count())->release();
				use_count_ = NULL;
#ifdef ASSIGNMENT_EXCEPTION
				// cast failed because of invalid assignment type
				throw Mp7JrsPtrException(MP7JRS_ERROR, "Invalid cast. Attempted to assign invalid pointer type.");
#endif
			}
		} else {
			// p is a null pointer, so are we
			ptr = NULL;
			use_count_ = NULL;
		}
	}

	~Mp7JrsPtr()
	{
		if (use_count_) {
			use_count_->release();
		}
	}

	template<class W>
	Mp7JrsPtr<P>& operator= (const Mp7JrsPtr<W>& p)
	{
		if (p != *this) {
			// Implement assignment in terms of construction
			Mp7JrsPtr<P> temp(p);
			swap(temp);
		}
		return *this;
	}

	// overwrite default assignment operator -- a template assignment
	// operator is never the default assignment operator, even if its
	// arguments would match
	// VC++6 needs this after the template operator=
	Mp7JrsPtr<P>& operator= (const Mp7JrsPtr<P>& p)
	{
		if (p != *this) {
			// Implement assignment in terms of construction
			Mp7JrsPtr<P> temp(p);
			swap(temp);
		}
		return *this;
	}

	template<class W>
	bool operator== (const Mp7JrsPtr<W>& p) const
	{
		return ptr == p.Ptr();
	}

	template<class W>
	bool operator< (const Mp7JrsPtr<W>& p) const
	{
		return ptr < p.Ptr();
	}

	template<class W>
	bool operator> (const Mp7JrsPtr<W>& p) const
	{
		return ptr > p.Ptr();
	}

	template<class W>
	bool operator!= (const Mp7JrsPtr<W>& p) const
	{
		return !operator==(p);
	}

	// enables  if (!smartptr) ...
	bool operator! () const
	{
		return ptr == NULL;
	}

	// enables  if (smartptr == 0)
	inline friend bool operator==(const Mp7JrsPtr<P> &lhs, const P *rhs)
	{
		return lhs.ptr == rhs;
	}

	// enables  if (0 == smartptr)
	inline friend bool operator==(const P *lhs, const Mp7JrsPtr<P> &rhs)
	{
		return lhs == rhs.ptr;
	}

	// enables  if (smartptr != 0)
	inline friend bool operator!=(const Mp7JrsPtr<P> &lhs, const P *rhs)
	{
		return lhs.ptr != rhs;
	}

	// enables  if (0 != smartptr)
	inline friend bool operator!=(const P *lhs, const Mp7JrsPtr<P> &rhs)
	{
		return lhs != rhs.ptr;
	}

	// TODO: perhaps write templated versions of the above four operators,
	// so that Smart Ptrs can be compared against normal pointers of
	// another type.

	// See: Alexei Alexandrescu, "Modern C++ Design", p. 178
private:
	class Tester
	{
		void operator delete(void *);
	};

public:
	// enables  if (smartptr)
	operator Tester*() const
	{
		if (!ptr) return 0;
		static Tester test;
		return &test;
	}

	P& operator*() const
	{
		if (ptr)
			return *ptr;
		else
			throw Mp7JrsPtrException(MP7JRS_FATAL_ERROR, "Attempted to dereference null pointer.");
	}

	P* operator->() const
	{
		if (ptr)
			return ptr;
		else
			throw Mp7JrsPtrException(MP7JRS_FATAL_ERROR, "Attempted to dereference null pointer.");
	}

	inline P* Ptr() const
	{
		return ptr;
	}

	inline Mp7Jrscounter * const *__count() const
	{
		// HACK - not for external use
		return &use_count_;
	}

	inline Mp7Jrscounter **__count()
	{
		// HACK - not for external use
		return &use_count_;
	}

/*
	// Auto-conversion is evil, but necessary so Smart Ptrs are
	// transparent to old extensions expecting raw pointers
	// TODO: can this go now?
	inline operator P *() const
	{
		return Ptr();
	}
*/

private:

	void swap(Mp7JrsPtr<P>& p)
	{
		P *temp = ptr;
		ptr = p.ptr;
		p.ptr = temp;
		Mp7Jrscounter *tmp = use_count_;
		use_count_ = p.use_count_;
		p.use_count_ = tmp;
	}

private:
	P *ptr;
	Mp7Jrscounter *use_count_;
};


template<class P>
class Mp7JrsWeakPtr {
public:

	Mp7JrsWeakPtr()
		: ptr(0), use_count_(0)
	{}

	template<class W>
	Mp7JrsWeakPtr(const Mp7JrsPtr<W>& p)
	{
		if (*(p.__count())) {
			(*p.__count())->add_ref();
			// It is now safe to access p.Ptr()
			ptr = dynamic_cast<P *>(p.Ptr());
			if (ptr) {
				use_count_ = const_cast<Mp7Jrscounter *>(*(p.__count()));
				use_count_->weak_add_ref();
				use_count_->release(); // release add_ref() from above
			}
			else {
				// Couldn't convert p.Ptr() to our type -> drop reference
				(*p.__count())->release();
				use_count_ = NULL;
#ifdef ASSIGNMENT_EXCEPTION
				// cast failed because of invalid assignment type
				throw Mp7JrsPtrException(MP7JRS_ERROR, "Invalid cast. Attempted to assign invalid pointer type.");
#endif
			}
		} else {
			// p is a null pointer, so are we
			ptr = NULL;
			use_count_ = NULL;
		}
	}

	Mp7JrsWeakPtr(const Mp7JrsWeakPtr<P>& p)
	{
		Mp7JrsPtr<P> temp(p.getPointer());
		// It is now safe to access the pointer.
		use_count_ = *(temp.__count());
		ptr = p.Ptr();
		if (use_count_) use_count_->weak_add_ref();
	}

	~Mp7JrsWeakPtr()
	{
		if (use_count_) {
			use_count_->weak_release();
		}
	}

	bool expired() const
	{
		if (use_count_)
			return use_count_->use_count() <= 0;
		else
			return true;
	}

	Mp7JrsPtr<P> getPointer() const
	{
		if (expired())
			return Mp7JrsPtr<P>();

		try {
			return Mp7JrsPtr<P>(*this);
		} catch (Mp7JrsPtrException const &) {
			return Mp7JrsPtr<P>();
		}
	}

	template<class W>
	Mp7JrsWeakPtr<P>& operator= (const Mp7JrsPtr<W>& p)
	{
		// Implement assignment in terms of construction
		Mp7JrsWeakPtr<P> temp(p);
		swap(temp);
		return *this;
	}

	template<class W>
	bool operator== (const Mp7JrsPtr<W>& p) const
	{
		return ptr == p.Ptr();
	}

	template<class W>
	bool operator< (const Mp7JrsPtr<W>& p) const
	{
		return ptr < p.Ptr();
	}

	template<class W>
	bool operator> (const Mp7JrsPtr<W>& p) const
	{
		return ptr > p.Ptr();
	}

	template<class W>
	bool operator!= (const Mp7JrsPtr<W>& p) const
	{
		return !operator==(p);
	}

	template<class W>
	bool operator== (const Mp7JrsWeakPtr<W>& p) const
	{
		return ptr == p.Ptr();
	}

	template<class W>
	bool operator< (const Mp7JrsWeakPtr<W>& p) const
	{
		return ptr < p.Ptr();
	}

	template<class W>
	bool operator> (const Mp7JrsWeakPtr<W>& p) const
	{
		return ptr > p.Ptr();
	}

	template<class W>
	bool operator!= (const Mp7JrsWeakPtr<W>& p) const
	{
		return !operator==(p);
	}


	inline P* Ptr() const
	{
		// Unsafe at any speed
		return ptr;
	}

	inline Mp7Jrscounter * const *__count() const
	{
		// HACK - not for external use
		return &use_count_;
	}

	inline Mp7Jrscounter **__count()
	{
		// HACK - not for external use
		return &use_count_;
	}

/*
	// Auto-conversion is evil, but necessary so Smart Ptrs are
	// transparent to old extensions expecting raw pointers
	// TODO: can this go now?
	inline operator P *() const
	{
		return Ptr();
	}
*/

private:

	void swap(Mp7JrsWeakPtr<P>& p)
	{
		P *temp = ptr;
		ptr = p.ptr;
		p.ptr = temp;
		Mp7Jrscounter *tmp = use_count_;
		use_count_ = *(p.__count());
		p.use_count_ = tmp;
	}

private:
	P *ptr;
	Mp7Jrscounter *use_count_;
};

#endif // MP7JRS_POINTER_INCLUDED
