/*****************************************************************************/
/* Copyright (C) 2003-2012, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


#ifndef MP7JRS_XPATHPARSECONTEXT_H
#define MP7JRS_XPATHPARSECONTEXT_H


#include <xercesc/util/XMLString.hpp>
#include "Mp7JrsUtil.h"


#include "Mp7JrsDefines.h"
#include "Mp7JrsNode.h"
#include "Mp7JrsEnumerator.h"
#include "Mp7JrsValueVectorOf.h" 

// Intro of API documentation
#include "docintro.h"


typedef enum {
	MP7JRS_XPATH_UNKNOWN = 0,
	
	MP7JRS_XPATH_ROOT, // /
	MP7JRS_XPATH_UP, // ../

	
	// The following operations are allowed for Get() and GetOrCreate()
	MP7JRS_XPATH_SELECT_ALL, // 
	MP7JRS_XPATH_SELECT_TYPED_ALL, // [@xsi:type='bla']
	
	MP7JRS_XPATH_SELECT_INDEX, // [i]
	MP7JRS_XPATH_SELECT_INDEXED_TYPE, // [i][@xsi:type]
	MP7JRS_XPATH_SELECT_TYPED_INDEX, // [@xsi:type][i]
	
	MP7JRS_XPATH_SELECT_LAST, // [last()]
	MP7JRS_XPATH_SELECT_LAST_TYPE, // [last()][xsi:type]
	MP7JRS_XPATH_SELECT_TYPED_LAST, // [xsi:type][last()]

	// Force node creation if in create mode
	MP7JRS_XPATH_FORCE_CREATE, // [position()>last()] [position()>last()][xsi:type] [xsi:type][position()>last()]

	// The following operations are allowed for Get() only
	MP7JRS_XPATH_SELECT_POSITION_RANGE, // [position()>A | position()<B]
	MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE, // [position()>A | position()<B][xsi:type]
	MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE, // [xsi:type][position()>A | position()<B]

	MP7JRS_XPATH_SELECT_POSITION_LT, // [position()<A]
	MP7JRS_XPATH_SELECT_POSITION_LT_TYPE, // [position()<A][xsi:type]
	MP7JRS_XPATH_SELECT_TYPED_POSITION_LT, // [xsi:type][position()<A]

	MP7JRS_XPATH_SELECT_POSITION_GT, // [position()>A]
	MP7JRS_XPATH_SELECT_POSITION_GT_TYPE, // [position()>A][xsi:type]
	MP7JRS_XPATH_SELECT_TYPED_POSITION_GT, // [xsi:type][position()>A]
	
} Mp7JrsXPathOperation;

typedef Mp7JrsValueVectorOf<int> Mp7JrsIndexVector;


/** Mp7Jrs XPath parsing context. 
 * Deals with all XPath related activities.
 */
struct MP7JRS_EXPORT Mp7JrsXPathParseContext
{
	Mp7JrsXPathParseContext();
	virtual ~Mp7JrsXPathParseContext();
	static bool Initialise(Mp7JrsXPathParseContext & c, Mp7JrsXPathParseContext ** context, const XMLCh *xpath, Mp7JrsXPathCreation flag);

	// Main routine
	static void Parse(Mp7JrsXPathParseContext * context);

	// Parsing helpers
	static void Reset(Mp7JrsXPathParseContext * context);
	static bool ScanSlash(Mp7JrsXPathParseContext * context);
	static bool ScanUp(Mp7JrsXPathParseContext * context);
	static bool ScanTail(Mp7JrsXPathParseContext * context);
	static bool ScanElement(Mp7JrsXPathParseContext * context);
	static bool ScanIndex(Mp7JrsXPathParseContext * context);
	static bool ScanXsiType(Mp7JrsXPathParseContext * context);
	static bool ConvertIndex(int startpos, int endpos, int minval, Mp7JrsXPathParseContext * context, int & n);
	
	static Mp7JrsNodePtr CreateNode(Mp7JrsXPathParseContext * context, unsigned int defaulttype);

	// Simple element
	static void PreProcessSimpleElement(Mp7JrsXPathParseContext * context);
	static bool HandleSimpleElement(Mp7JrsXPathParseContext * context, Mp7JrsNodePtr element);
	
	
	// Collections
	static bool PreProcessCollection(Mp7JrsXPathParseContext * context, unsigned int collsize, int upperboundary);
	static int HandleCollection(Mp7JrsXPathParseContext * context, int i, Mp7JrsNodePtr element, Mp7JrsNodeEnum & result);
	static bool PostProcessElementCollection(Mp7JrsXPathParseContext * context, int & i, Mp7JrsNodeEnum & result);
	static bool PostProcessSequenceCollection(Mp7JrsXPathParseContext * context, int & i, int & start, int & stop, Mp7JrsNodeEnum & result);
	static bool PostProcessChoiceCollection(Mp7JrsXPathParseContext * context, int & i, Mp7JrsNodeEnum & result);
	
 
	// TODO Moving
	static Mp7JrsNodeEnum Root(Mp7JrsXPathParseContext * context, const Mp7JrsNodePtr node, Mp7JrsClientID client);
	static Mp7JrsNodeEnum Up(Mp7JrsXPathParseContext * context, const Mp7JrsNodePtr node, Mp7JrsClientID client);
	

	// Some exceptions
	static void ErrorBoundary(Mp7JrsXPathParseContext * context);
	static void ErrorUnallowedExpression(Mp7JrsXPathParseContext * context);
	static void ErrorAbstractType(Mp7JrsXPathParseContext * context);
	static void ErrorInvalidIndex(Mp7JrsXPathParseContext * context);
	static void ErrorInvalidEmptyItem(Mp7JrsXPathParseContext * context, int i);
	static void ErrorNoFillingGaps(Mp7JrsXPathParseContext * context, unsigned int lastvalidindex);
	static void ErrorCannotSetElement(Mp7JrsXPathParseContext * context);
	static void ErrorUnknownSubElement(Mp7JrsXPathParseContext * context, const char * type);
	static void ErrorNoSubElement(Mp7JrsXPathParseContext * context, const char * type);
	static void ErrorUnsupportedType(Mp7JrsXPathParseContext * context, const char * type);
	static void ErrorWrongCaller(const XMLCh * xpath);
	static void ErrorNotImplemented(const XMLCh * xpath, const char * type);
	
	// FTT Workaround for setting content data in XPath
	static bool ContentFromString(Mp7JrsNode * node, const XMLCh * const txt);
	// FTT Workaround for getting content data in XPath
	static XMLCh * ContentToString(const Mp7JrsNode * node);
	
	// FTT Using our locationmap for now, TODO: change the name to something like xsitype_to_typeid()
	// FTT Changed name from ResolveNamespace to
	static XMLCh * ResolveTypeName(const XMLCh * xsitype);
	

	Mp7JrsXPathOperation Operation;
	bool Found;
	XMLCh * ElementName;
	XMLCh * XsiType;
	const XMLCh * XPath; // Global, used for error messages
	XMLCh * UnparsedXPath;
	int StartIndex; // -3 = take all, -2 = take last
	int StopIndex;
	
	bool Create; // Global flag
	bool PreferType;
	unsigned int Id;
	int CollectionSize;
	int UpperBoundary;
	
	int ItemCount; // Used to count [i] ...
	int ValidItemCount; // For [xsi:type i] or [i xsi:type]
	int LastItemIndex;
	int LastValidItemIndex;
	Mp7JrsNodePtr LastNode;
	Mp7JrsIndexVector FreeIndexList;
	XMLCh * QualifiedXsiType; // Format "URI:TYPENAME"
	
};	

#endif // MP7JRSXPATHPARSECONTEXT_H
